(* $Id: hydro_client_connector.ml 15333 2007-11-23 16:41:36Z gerd $ *)

open Hydro_types
open Hydro_util


type socket =
    [ `Connect_addr of Unix.sockaddr * transport_protocol_type
    | `Connect_inet of string * int * transport_protocol_type
    | `Unconnected_datagram of Unix.socket_domain
    | `Descriptor of unit -> descriptor
    ]


type connection =
    < ctrl : Hydro_transport.hydro_multiplex_controller;
      shutdown : unit -> unit
    >


class type mplex_config =
object
  method multiplexing : 
    descriptor -> Unixqueue.event_system -> connection Uq_engines.engine
end


let mplex_of_descr descr esys =
  match descr # proto_type with
    | `Stream ->
	Hydro_transport.stream_hydro_multiplex_controller
	  ~close_inactive_descr:false  (* this is done by descr#shutdown *)
	  descr#file_descr 
	  esys

    | `Datagram ->
	Hydro_transport.datagram_hydro_multiplex_controller
	  ~close_inactive_descr:false  (* this is done by descr#shutdown *)
	  descr#file_descr 
	  esys


class default_mplex_config : mplex_config =
object
  method multiplexing descr esys =
    let eng =
      try
        let mplex = mplex_of_descr descr esys in
	let c =
	  ( object 
	      method ctrl = mplex
	      method shutdown() =
		mplex#inactivate();
		descr#shutdown()
	    end
	  ) in
        new Uq_engines.epsilon_engine (`Done c) esys
      with
        | error ->
            new Uq_engines.epsilon_engine (`Error error) esys in
    Uq_engines.when_state
      ~is_aborted:(fun () -> descr#shutdown())
      ~is_error:(fun _ -> descr#shutdown())
      eng;
    eng

end


let default_mplex_config = new default_mplex_config


type connector =
    [ `Mplex_controller of Hydro_transport.hydro_multiplex_controller
    | `Mplex_socket of socket * mplex_config
    | `Proxy_endpoint of endpoint * network_port option
    ]


let endpoint_reg = Hashtbl.create 10

let register_endpoint_type ept handler modelist =
  Hashtbl.replace endpoint_reg ept (handler, modelist)


let socket_type =
  function
    | `Stream -> Unix.SOCK_STREAM
    | `Datagram -> Unix.SOCK_DGRAM


let descriptor fd tpt =
  ( object
      val mutable is_down = false
      method file_descr = assert(not is_down); fd
      method proto_type = tpt
      method shutdown() =
	if not is_down then Unix.close fd;
	is_down <- true;
    end
  )

let connect_engine_for_ca ca tpt cfg esys =
  new Uq_engines.seq_engine
    (try 
       Uq_engines.connector ca esys
     with Not_found as nf ->    (* HACK *)
       match ca with
	 | `Socket(`Sock_inet_byname(_,host,_),_) ->
	     new Uq_engines.epsilon_engine
	       (`Error (Domain_not_found host))
	       esys
	 | _ -> raise nf
    )
    (fun cs ->
       let fd = Uq_engines.client_socket cs in
       let descr = descriptor fd tpt in
       cfg#multiplexing descr esys
    )


let socket_connect_engine sock cfg esys =
  match sock with
    | `Connect_addr(sa,tpt) ->
	let st = socket_type tpt in
	let ca =
	  `Socket ( (match sa with
		       | Unix.ADDR_INET(ia,p) -> `Sock_inet(st,ia,p)
		       | Unix.ADDR_UNIX path  -> `Sock_unix(st,path)
		    ),
		    Uq_engines.default_connect_options
		  ) in
	connect_engine_for_ca ca tpt cfg esys

    | `Connect_inet(host,port,tpt) ->
	let st = socket_type tpt in
	let ca =
	  `Socket ( `Sock_inet_byname(st,host,port),
		    Uq_engines.default_connect_options
		  ) in
	connect_engine_for_ca ca tpt cfg esys

    | `Unconnected_datagram dom ->
	let st = Unix.SOCK_DGRAM in
	let fd = Unix.socket dom st 0 in
	let descr = descriptor fd `Datagram in
	cfg#multiplexing descr esys

    | `Descriptor f ->
	let descr = f() in
	cfg#multiplexing descr esys


let connect_engine (connector : connector) esys =
  match connector with
    | `Mplex_controller ctrl ->
	let c =
	  (object method ctrl = ctrl method shutdown = ctrl#inactivate end) in
	new Uq_engines.epsilon_engine (`Done c) esys

    | `Mplex_socket(sock, cfg) ->
	socket_connect_engine sock cfg esys

    | `Proxy_endpoint (ep, port_opt) ->
	let ept = endpoint_type ep in
	let (sock, cfg) =
	  try
	    let handler, _ = Hashtbl.find endpoint_reg ept in
	    handler ep port_opt
	  with
	    | Not_found ->
		failwith("No handler registered for endpoint type " ^ 
			   string_of_endpoint_type ept) in
	socket_connect_engine sock cfg esys


(* Handlers for TCP and UDP: *)

let simple_handler ep port_opt =
  match ep with
    | `TCP tcp ->
	( match port_opt with
	    | Some(`TCP(inet,port)) ->
		(`Connect_addr(Unix.ADDR_INET(inet,port), `Stream),
		 default_mplex_config
		)
	    | Some _ ->
		failwith "Endpoint handler for TCP: Got non-TCP resolution"
	    | None ->
		(`Connect_inet(tcp#host, tcp#port, `Stream), 
		 default_mplex_config
		)
	)
    | `UDP udp ->
	( match port_opt with
	    | Some(`UDP(inet,port)) ->
		(`Connect_addr(Unix.ADDR_INET(inet,port), `Datagram),
		 default_mplex_config
		)
	    | Some _ ->
		failwith "Endpoint handler for UDP: Got non-UDP resolution"
	    | None ->
		(`Connect_inet(udp#host, udp#port, `Datagram), 
		 default_mplex_config
		)
	)
    | _ -> assert false

let () =
  register_endpoint_type
    `TCP
    simple_handler
    [ `Twoway; `Oneway; `Batch_oneway ];
  register_endpoint_type 
    `UDP
    simple_handler
    [ `Datagram; `Batch_datagram ]

let proxy_supports_endpoint_type ept pm =
  try
    let _, pmlist =
      Hashtbl.find endpoint_reg ept in
    List.mem pm pmlist
  with
    | Not_found -> false

