(* $Id: hydro_endpoint.ml 15339 2007-11-23 19:24:42Z gerd $ *)


(* This is an adaption of Ocamlnet's rpc_client.ml *)

open Hydro_types
open Hydro_util
open Hydro_dbg

module SessionMap = Map.Make(Int32)

type client_response_condition = [ `Success | client_condition ]

type client_response =
    < condition : client_response_condition;
      out_args : value array;
      result : value;
         (* Raises [Client_condition] if [condition] is not [`Success] *)
      peer : Unix.sockaddr option;
      mode : op_mode;
      client : client option
    >


and call_state =
    Waiting    (* call has not yet been sent *)
  | Pending    (* call has been sent, still no answer *)
  | Done       (* got answer for the call *)


and call_msg =
    [ `Request of request_msg
    | `Batch_request of batch_request_msg list
    ]


and cl_state =
  | Init
  | Down
  | Connecting of Hydro_client_connector.connection Uq_engines.engine
  | Validating
  | Closing1    (* close message not yet sent *)
  | Closing2    (* close message is sent out *)
  | Up

and call =
      { msg : call_msg;
	hf : hfunction;
	
	(* Params: *)
        destination : Unix.sockaddr option;
	msg_timeout : float;
	mutable msg_timeout_group : Unixqueue.group option;
	
        pass_response : client_response -> unit;

        mutable state : call_state;
	mutable peer : Unix.sockaddr option;

        (* mutable when_sent : unit -> bool; *)
      }

and client =
    { ci : < >;    (* client identifier *)
      sys : system;

      mutable trans : Hydro_client_connector.connection option;
      esys : Unixqueue.event_system;

      mutable cl_state : cl_state;

      mutable wait_queue : call Queue.t;
      mutable wait_map : call SessionMap.t;
      mutable pending : call SessionMap.t;

      mutable next_xid : int32;

      mutable client_params : client_params;  (* as configured *)
      mutable max_proto_minor : int;
      mutable max_enc_minor : int;

      mutable idle_timeout_group : Unixqueue.group option;
      mutable shutdown_when_idle : bool;

      mutable shutdown_timeout_group : Unixqueue.group option;

      mutable pool_callback : (unit -> unit) option;
      mutable graceful : bool;
    }

let xidopt_of_call call =
  match call.msg with
    | `Request req ->
	let xid = req#request_id in
	if xid = 0l then
	  None
	else
	  Some xid
    | _ ->
	None

let max_proto_minor cl trans =
  (* Get max protocol version of the server *)
  match trans#ctrl#transport_protocol_type with
    | `Stream ->
	cl.max_proto_minor
    | `Datagram ->
	( match cl.client_params#max_proto_minor with
	    | None -> 0  (* conservative default *)
	    | Some pm -> pm
	)


let max_enc_minor cl trans =
  (* Get max encoding version of the server *)
  match trans#ctrl#transport_protocol_type with
    | `Stream ->
	cl.max_enc_minor
    | `Datagram ->
	( match cl.client_params#max_enc_minor with
	    | None -> 0  (* conservative default *)
	    | Some em -> em
	)

let max_enc_minor_nt cl =
  match cl.trans with
    | Some trans -> max_enc_minor cl trans
    | None -> assert false


  (*****)

let check_for_input =  (* "forward declaration" *)
  ref (fun _ -> ())

let check_for_output = (* "forward declaration" *)
  ref (fun _ -> ())

  (*****)

let stop_call_timers cl call =
  match call.msg_timeout_group with
    | None -> ()
    | Some g ->
	Unixqueue.clear cl.esys g;
	call.msg_timeout_group <- None


let pass_response cl call response =
  if call.state <> Done then (  (* Don't call back twice *)
    (* Stop timers, if any : *)
    stop_call_timers cl call;
    (* Change the state of the call to 'Done': *)
    call.state <- Done;
    (* pass [response] to the call back function: *)
    try
      dlog "Hydro_client: Calling back";
      call.pass_response response;
      dlog "Hydro_client: Returned from callback";
    with
      | Unbound_exception x ->
          dlog "Hydro_client: Unbound_exception";
          raise x
      | any ->
          begin  (* pass the exception to the exception handler: *)
	    dlogf
              "Hydro_client: Exception from callback: %s"
              (Hydro_util.exn_to_string any);
            cl.client_params#exception_handler#handle any
          end
  )

let pass_condition cl call cc connect_flag =
  (* Caution! This function does not remove [call] from the set of pending
   * calls.
   *)
  let cc =
    (* Check for connect errors: *)
    match cc with
      | (`Message_timeout | `Transport_timeout) when connect_flag ->
	  `Connect_timeout
      | `Error x when connect_flag -> 
	  `Connect_error x
      | `Message_lost _ ->
	  `Message_lost (not connect_flag && call.state <> Waiting)
      | _ -> cc in
  dlog ("Hydro_client: Passing condition " ^ cc_to_string cc);
  let response =
    ( object
	method condition = (cc : client_condition :> client_response_condition)
	method out_args = [| |]
	method result = raise (Client_condition cc)
	method peer = call.peer
	method mode = call.hf#mode
	method client = Some cl
      end
    ) in
  pass_response cl call response

let pass_condition_to_all cl cc connect_flag =
  (* Caution! This function does not erase the set of pending calls.  *)
  dlog "Hydro_client: Passing condition to all";

  let pending = cl.pending in
  cl.pending <- SessionMap.empty;

  let waiting = Queue.create() in
  Queue.transfer cl.wait_queue waiting;
  cl.wait_map <- SessionMap.empty;

  SessionMap.iter 
    (fun _ call -> pass_condition cl call cc connect_flag) 
    pending;
  Queue.iter
    (fun call -> pass_condition cl call cc connect_flag) 
    waiting

  (*****)

let is_up cl =
  match cl.cl_state with
    | Init
    | Up
    | Validating
    | Connecting _ ->
	true
    | _ ->
	false

let graceful cl =
  cl.graceful

let tell_pool cl =
  match cl.pool_callback with
    | None -> ()
    | Some f -> f()


let stop_shutdown_timer cl =
  match cl.shutdown_timeout_group with
    | None -> ()
    | Some g -> 
	Unixqueue.clear cl.esys g;
	cl.shutdown_timeout_group <- None


let abort_with cc_opt cl =
  if cl.cl_state <> Down then (  (* Avoid nested aborts *)
    dlog "Hydro_client: Aborting";
    if cc_opt <> None then
      cl.graceful <- false;
    stop_shutdown_timer cl;
    let was_up = is_up cl in
    let connect_flag =
      match cl.cl_state with
	| Connecting _ -> true
	| _ -> false in
    ( match cl.cl_state with
	| Down -> ()

	| Connecting e ->
	    e#abort()
	      
	| Validating | Closing1 | Closing2 | Up ->
	    ( match cl.trans with
		| None -> () (* strange *)
		| Some c ->
		    c # shutdown();
		    cl.trans <- None
	    )
	      
	| Init ->
	    ()
    );
    cl.cl_state <- Down;
    let cc =
      match cc_opt with
	| Some cc -> cc
	| None -> `Message_lost false in  (* arg updated later *)
    if was_up then 
      tell_pool cl;
    pass_condition_to_all cl cc connect_flag;
  )



let queued_requests cl =
  match cl.cl_state with
    | Down -> 
	0
    | _ ->
	let n_pending =
	  SessionMap.fold (fun _ _ n -> n+1) cl.pending 0 in
	Queue.length cl.wait_queue + n_pending

let is_idle cl =
  cl.cl_state = Up &&
    cl.wait_map = SessionMap.empty &&
    cl.pending = SessionMap.empty
  (* Note: cl.wait_queue is unreliable for checking on idleness. The
     queue may contain cancelled calls.
   *)


let client_id cl =
  Oo.id cl.ci


let start_shutdown_timer cl =
  if cl.client_params#msg_timeout >= 0.0 then (
    (* If there is a msg_timeout this is also the upper bound for the 
       shutdown process. Note that a trans_timeout also applies because
       it is still set on the transporter.
     *)
    let g = Unixqueue.new_group cl.esys in
    Unixqueue.once cl.esys g cl.client_params#msg_timeout
      (fun () ->
	 (* Transport_timeout seems to be most appropriate *)
	 abort_with (Some `Transport_timeout) cl
      );
    cl.shutdown_timeout_group <- Some g
  )


let shutdown cl =
  match cl.cl_state with
    | Up ->
	( match cl.trans with
	    | None -> assert false
	    | Some trans ->
		if trans#ctrl#transport_protocol_type = `Stream then (
		  dlog "Hydro_client: Will send shutdown msg";
		  start_shutdown_timer cl;
		  cl.cl_state <- Closing1;
		  tell_pool cl;
		  !check_for_output cl
		) else 
		  (* Close message isn't supported for datagrams: *)
		  abort_with None cl
	)

    | Init
    | Validating
    | Connecting _ ->
	abort_with None cl

    | Closing1
    | Closing2 ->
	(* Shutdown already in progress *)
	()

    | Down ->
	(* ignore *)
	()
	

let shutdown_when_idle cl =
  match cl.cl_state with
    | Init
    | Up
    | Validating
    | Connecting _ ->
	cl.shutdown_when_idle <- true
    | _ ->
	()

let abort cl =
  abort_with None cl

  (*****)

let cancel_idle_timeout cl =
  match cl.idle_timeout_group with
    | None -> ()
    | Some g ->
	Unixqueue.clear cl.esys g;
	cl.idle_timeout_group <- None
	  
let when_idle cl =
  cancel_idle_timeout cl;
  if cl.shutdown_when_idle then
    shutdown cl
  else (
    let tmo = cl.client_params#idle_timeout in
    if tmo > 0.0 then (
      let g = Unixqueue.new_group cl.esys in
      Unixqueue.once cl.esys g tmo
	(fun () -> 
	   cl.idle_timeout_group <- None;
	   shutdown cl
	);
      cl.idle_timeout_group <- Some g;
    )
  )

  (*****)

let remove_call cl call =
  ( match xidopt_of_call call with
      | Some xid ->
	  cl.pending <- SessionMap.remove xid cl.pending;
	  cl.wait_map <- SessionMap.remove xid cl.wait_map;
      | None ->
	  (* No reply expected... *)
	  ()
  )
  (* Note that we do not remove the call from waiting_calls for
     performance reasons. We simply skip it there if we find it.
     pass_exception will set call.state to Done.
   *)

let cancel_call cl call cc =
  dlog "Hydro_client: Cancelling call";
  remove_call cl call;
  let connect_flag =
    match cl.cl_state with
      | Connecting _ -> true
      | _ -> false in
  pass_condition cl call cc connect_flag;
  if is_idle cl then when_idle cl;
  !check_for_output cl

let set_timeout cl call =
  if call.msg_timeout > 0.0 && call.msg_timeout_group = None then (
    (* Note: Case call_timeout = 0.0 is handled elsewhere *)
    let g = Unixqueue.new_group cl.esys in
    Unixqueue.once cl.esys g call.msg_timeout
      (fun () ->
         call.msg_timeout_group <- None;
         dlog "Hydro_client: Timeout handler";
         cancel_call cl call `Message_timeout;
         (* Maybe we have to cancel reading: *)
         !check_for_input cl
      );
    call.msg_timeout_group <- Some g
  )

let rec new_xid cl =
  let xid = cl.next_xid in
  cl.next_xid <- Int32.succ xid;
  if (xid = 0l ||     (* 0 is reserved for oneway calls *)
      SessionMap.mem xid cl.wait_map || 
      SessionMap.mem xid cl.pending)
  then
    new_xid cl
  else
    xid


let find_function intf op =
  try List.find (fun f -> f#name = op) intf#elements
  with Not_found -> raise(Marshal_error ("Operation not found: " ^ op))


let add_call cl msg hf call_params pass_response =
  cancel_idle_timeout cl;
  let msg_timeout =
    match call_params#msg_timeout with
      | Some tmo -> tmo
      | None -> cl.client_params#msg_timeout in
  let call =
    { msg = msg;
      hf = hf;
      destination = call_params#destination;
      msg_timeout = msg_timeout;
      msg_timeout_group = None;
      pass_response = pass_response;
      state = Waiting;
      peer = None
    } in
  if msg_timeout > 0.0 then
    set_timeout cl call;
  Queue.add call cl.wait_queue;
  ( match xidopt_of_call call with
      | None -> ()
      | Some xid ->
	  cl.wait_map <- SessionMap.add xid call cl.wait_map
  );
  !check_for_output cl


let create_request facet cl id hf op call_params in_args xid =
  let msg = 
    ( object
	method request_id = xid
	method id = id
	method facet = facet
	method operation = op
	method mode = hf#mode;
	method context = [| |]     (* TODO *)
	method params =
	  (* We may only call this method when we have trans! *)
	  let em =
	    min Hydro_marshal.max_enc_minor (max_enc_minor_nt cl) in
	  let in_type =
	    `Struct hf#in_args in
	  let cflag =
	    hf # in_classes in
	  let b = 
	    Netbuffer.create 500 in
	  Hydro_marshal.marshal cl.sys in_type cflag (`Struct in_args) b em;
	  { encap_buf = b;
	    encap_pos = 0;
	    encap_len = Netbuffer.length b;
	    encap_enc_minor = em
	  }
      end
    ) in
  `Request msg

let require_client_is_up cl =
  match cl.cl_state with
    | Down | Closing1 | Closing2 ->
	raise(Client_condition `Client_is_down)
    | _ ->
	()

let twoway_call ?facet cl id (intf:hintf) op (call_params:call_params) args pass_response =
  require_client_is_up cl;
  let hf = find_function intf op in
  let xid = new_xid cl in
  let msg = create_request facet cl id hf op call_params args xid in
  add_call cl msg hf call_params pass_response

  (*****)

let decode_reply cl call reply peer_opt =
  match reply#result with
    | `Success eb ->
	let hf = call.hf in
	let t =
	  `Struct
	    [| (* Contrary to the documentation, the result value comes
                  first, followed by the out arguments!
                *)
	      "resval", hf#result;
	      "out_args", `Struct hf#out_args;
	    |] in
	let cflag = hf#out_classes in
	let v = Hydro_unmarshal.unmarshal cl.sys t cflag eb in
	( match v with
	    | `Struct [| resval;
			 `Struct out_vals;
		      |] ->
		( object 
		    method condition = `Success
		    method out_args = out_vals
		    method result = resval
		    method peer = peer_opt
		    method mode = call.hf#mode
		    method client = Some cl
		  end
		)
	    | _ ->
		assert false
	)

    | `User_exception eb ->
	let slices = Hydro_unmarshal.unmarshal_exn cl.sys eb in
	let cc = `User_exception slices in
	( object
	    method condition = cc
	    method out_args = [| |]
	    method result = raise(Client_condition cc)
	    method peer = peer_opt
	    method mode = call.hf#mode
	    method client = Some cl
	  end
	)
	
    | ( `Object_does_not_exist _
      | `Facet_does_not_exist _
      | `Operation_does_not_exist _
      | `Unknown_local_exception _
      | `Unknown_user_exception _
      | `Unknown_exception _ as cc
      ) ->
	( object 
	    method condition = (cc :> client_response_condition)
	    method out_args = [| |]
	    method result = raise(Client_condition cc)
	    method peer = peer_opt
	    method mode = call.hf#mode
	    method client = Some cl
	  end
	)

  (*****)

let process_incoming_message cl hdr nb peer_opt =
  (* First decode the message: *)
  let response_opt =
    try
      let eb =
	{ encap_buf = nb;
	  encap_pos = 0;
	  encap_len = Netbuffer.length nb;
	  encap_enc_minor = hdr#enc_minor
	} in
      let msg = Hydro_unmarshal.unmarshal_msg cl.sys hdr eb in
      (* checks version in hdr *)
      
      ( match msg with
	  | `Reply reply -> 
	      dlog "Hydro_client: Got reply";
	      let xid = reply#request_id in
	      let call_opt =
		try Some(SessionMap.find xid cl.pending)
		with Not_found -> None in
	      ( match call_opt with
		  | Some call ->
		      assert(call.state = Pending);
		      let response = decode_reply cl call reply peer_opt in
		      remove_call cl call;
		      Some (call, response)
		    
		  | None ->
		      (* This could be a timed-out reply. Simply ignore! *)
		      None
	      )
		
	  | `Validate_connection ->
	      dlog "Hydro_client: Got Validate_connection";
	      ( match cl.cl_state with
		  | Validating ->
		      cl.cl_state <- Up;
		      cl.max_proto_minor <- hdr#proto_minor;
		      cl.max_enc_minor <- hdr#enc_minor;
		  | Up -> 
		      ()
		  | _ ->
		      assert false
	      );
	      None

	  | `Close_connection ->
	      dlog "Hydro_client: Got Close_connection";
	      let was_up = is_up cl in
	      ( match cl.cl_state with
		  | Validating
		  | Up 
		  | Closing1 ->
		      ( match cl.trans with
			  | None -> assert false
			  | Some trans ->
			      if trans#ctrl#transport_protocol_type = `Stream 
			      then
				cl.cl_state <- Closing2
			(* Close message is ignored for datagrams *)
		      );
		      if was_up then tell_pool cl
		  | Closing2 -> 
		    ()
		  | _ ->
		      assert false
	      );
	      None

	  | _ ->
	      raise(Protocol_violation `BadMessageType)
      )
    with
      | err ->
	  (* An exception that is not bound to a specific call: *)
	  dlog ("Hydro_client: Exception decoding reply: " ^
                  Hydro_util.exn_to_string err);
          cl.client_params#exception_handler#handle err;
	  None
  in
  match response_opt with
    | Some (call,response) ->
	pass_response cl call response
    | None ->
	()

  (*****)

let rec handle_incoming_message cl r =
  (* Called when a complete message has been read by the transporter *)
  match r with
    | `Error err ->
	abort_with (Some (`Error err)) cl

    | `Ok(hdr, nb, addr) ->
	dlog "Hydro_client: Message arrived";
	let peer_opt =
	  match addr with
	    | `Implied -> None 
	    | `Sockaddr a -> Some a in
	let prev_state = cl.cl_state in
	process_incoming_message cl hdr nb peer_opt;
	let is_validated = (prev_state = Validating) && (cl.cl_state = Up) in
	if is_idle cl then
	  when_idle cl;
	next_incoming_message cl;
	if is_validated then !check_for_output cl;
	()

    | `End_of_file ->
	dlog "Hydro_client: End of file";
	abort_with None cl

and next_incoming_message cl =
  match cl.trans with
    | None ->
	(* Not yet connected *)
	()

    | Some trans ->
	trans # ctrl # cancel_rd_polling();
	(* TODO: A better criterion for controlling reading *)
	let do_read =
	  match cl.cl_state with
	    | Init -> assert false
	    | Down -> false
	    | Connecting _ -> false
	    | Validating -> true
	    | Up ->
		cl.pending <> SessionMap.empty || trans#ctrl#writing
		  (* That means: we read when we expect a reply, or
                     when we are writing currently
                   *)
	    | Closing1 -> true
	    | Closing2 -> true in
	if not trans#ctrl#reading && do_read then (
	  dlog "Hydro_client: (re)starting reading";
	  trans # ctrl # start_reading
	    ~when_done:(fun r -> handle_incoming_message cl r)
	    ()
	)
;;


check_for_input := next_incoming_message ;;


let rec handle_outgoing_message cl call r =
  (* Called after a complete message has been sent by the transporter *)
  match r with
    | `Error err ->
	abort_with (Some (`Error err)) cl

    | `Ok () ->
	dlog "Hydro_client: message writing finished";
	!check_for_input cl;
	next_outgoing_message cl

and next_outgoing_message cl =
  match cl.trans with
    | None -> ()
    | Some trans ->
	if not trans#ctrl#writing then (
	  match cl.cl_state with
	    | Init         -> assert false
	    | Down         -> ()
	    | Connecting _ -> assert false   (* where is [trans] from? *)
	    | Validating   -> ()   (* not yet! *)
	    | Up           -> next_outgoing_call cl trans
	    | Closing1     -> send_close_msg cl trans
	    | Closing2     -> ()     (* no more output! *)
	)

and next_outgoing_call cl trans =
  let call_opt =
    try Some(Queue.take cl.wait_queue) with Queue.Empty -> None in

  match call_opt with
    | Some call when call.state = Waiting ->
	(* Remove it also from the wait_map, if there is a xid. Add it
           to the pending calls.
	 *)
	( match xidopt_of_call call with
	    | None -> ()
	    | Some xid ->
		cl.wait_map <- SessionMap.remove xid cl.wait_map;
		cl.pending  <- SessionMap.add xid call cl.pending;
		call.state  <- Pending;
		set_timeout cl call;
	);
	(* Send the message: *)
	let pm =
	  min Hydro_marshal.max_proto_minor (max_proto_minor cl trans) in
	let em =
	  min Hydro_marshal.max_enc_minor (max_enc_minor cl trans) in
	let mb = 
	  Hydro_marshal.marshal_msg 
	    cl.sys `Compression_unsupported  (call.msg :> msg) pm em in
	let dest =
	  match call.destination with
	    | Some d -> `Sockaddr d
	    | None -> trans#ctrl#getpeername in
	dlog "Hydro_client: start_writing";
        trans # ctrl # start_writing
          ~when_done:(fun r ->
                        handle_outgoing_message cl call r)
	  mb
          dest;
	!check_for_input cl

    | Some _ ->
	(* The call was cancelled before issued! *)
	dlog "Hydro_client: found cancelled call"

    | None -> 
	()

and send_close_msg cl trans =
  let pm =
    min Hydro_marshal.max_proto_minor (max_proto_minor cl trans) in
  let em =
    min Hydro_marshal.max_enc_minor (max_enc_minor cl trans) in
  let mb = 
    Hydro_marshal.marshal_msg cl.sys `Compression_unsupported `Close_connection pm em in
  let dest = trans#ctrl#getpeername in
  dlog "Hydro_client: start_writing close msg";
  trans # ctrl # start_writing
    ~when_done:(fun r -> handle_close_msg cl trans r)
    mb
    dest


and handle_close_msg cl trans r =
  (* Called after a close message has been sent by the transporter *)
  match r with
    | `Error err ->
	abort_with (Some (`Error err)) cl

    | `Ok () ->
	dlog "Hydro_client: close msg writing finished";
	trans#ctrl#start_writing_eof
	  ~when_done:(fun r -> handle_out_eof cl trans r)
	  ()

	  (* No more next message! *)

and handle_out_eof cl trans r =
  (* Called after an EOF has been sent by the transporter *)
  match r with
    | `Error err ->
	abort_with (Some (`Error err)) cl

    | `Ok () ->
	dlog "Hydro_client: EOF sent";
	cl.cl_state <- Closing2;
	!check_for_input cl

	  (* No more next message! *)
;;


check_for_output := next_outgoing_message ;;


  (*****)

let default_client_params =
  Hydro_params.client_params()


let create_client sys conn esys =
  let cl =
    { ci = (object end);
      sys = sys;
      trans = None;
      esys = esys;
      cl_state = Init;
      wait_queue = Queue.create();
      wait_map = SessionMap.empty;
      pending = SessionMap.empty;
      next_xid = 1l;
      client_params = default_client_params;
      max_proto_minor = 0;  (* be conservative *)
      max_enc_minor = 0;    (* be conservative *)
      shutdown_when_idle = false;
      idle_timeout_group = None;
      shutdown_timeout_group = None;
      pool_callback = None;
      graceful = true;
    } in

  (* Start connecting a bit later: *)
  let cg = Unixqueue.new_group esys in
  Unixqueue.once esys cg 0.0
    (fun () ->
       if cl.cl_state = Init then (
	 (* i.e. not shut down in the meantime *)

	 let ce =
	   Hydro_client_connector.connect_engine conn esys in
	 cl.cl_state <- Connecting ce;
	 
	 (* Now somebody might have configured the client for a trans_timeout: *)
	 if cl.client_params#trans_timeout >= 0.0 then (
	   Unixqueue.once esys cg cl.client_params#trans_timeout
	     (fun () -> 
		abort_with
		  (Some `Connect_timeout)
		  cl
	     )
	 );
	 
	 Uq_engines.when_state
	   ~is_done:(fun c ->
		       dlog "Hydro_client: connected";
		       Unixqueue.clear esys cg;
		       match cl.cl_state with
			 | Connecting _ ->
			     cl.trans <- Some c;
			     if cl.client_params#trans_timeout >= 0.0 then (
			       dlog "Hydro_client: setting trans_timeout";
			       c # ctrl # set_timeout 
				 ~notify:(fun () -> 
					    abort_with (Some `Transport_timeout) cl
					 )
				 cl.client_params#trans_timeout;
			     );
			     ( match c#ctrl#transport_protocol_type with
				 | `Stream ->
				     cl.cl_state <- Validating;
				     !check_for_input cl
				 | `Datagram ->
				     cl.cl_state <- Up;
				     !check_for_input cl;
				     !check_for_output cl
			     )
			 | Down -> 
			     ()  (* strange *)
			 | _ ->
			     assert false
		    )
	   ~is_error:(fun err ->
			Unixqueue.clear esys cg;
			abort_with (Some (`Connect_error err)) cl;
		     )
	   ~is_aborted:(fun () ->
			  Unixqueue.clear esys cg
		       )
	   ce;
       )
    );

  cl


let configure_client cl params =
  if cl.pool_callback <> None then
    failwith "Hydro_client.configure: Cannot reconfigure pool clients";
  cl.client_params <- params


let pool_connect cl f =
  if cl.pool_callback <> None then
    failwith "Hydro_client.pool_connect: already pool member";
  cl.pool_callback <- Some f  



module Client = struct
  (* The interface for clients *)

  type t = client

  let create = create_client

  let configure = configure_client

  type response = client_response

  let twoway_call = twoway_call

  let queued_requests = queued_requests

  let is_up = is_up

  let is_idle = is_idle

  let client_id = client_id

  let shutdown = shutdown

  let shutdown_when_idle = shutdown_when_idle

  let abort = abort

  let graceful = graceful

  let pool_connect = pool_connect

end

